var EXPORTED_SYMBOLS = ["GM_API", "GM_apiSafeCallback", "GM_waiveXrays", "GM_checkedPrincipal"];

const Cu = Components.utils;
Cu.import("resource://scriptish/constants.js");

lazyImport(this, "resource://scriptish/logging.js", ["Scriptish_logError", "Scriptish_logScriptError", "Scriptish_log"]);

const { Style } = jetpack("sdk/stylesheet/style");
const { attach, detach } = jetpack("sdk/content/mod");

lazyUtil(this, "cryptoHash");
lazyUtil(this, "getScriptHeader");
lazyUtil(this, "notification");
lazyUtil(this, "openInTab");
lazyUtil(this, "stringBundle");

lazyImport(this, "resource://scriptish/api/GM_ScriptStorage.js", ["GM_ScriptStorage"]);
lazyImport(this, "resource://scriptish/api/GM_xmlhttpRequester.js", ["GM_xmlhttpRequester"]);
lazyImport(this, "resource://scriptish/api/GM_Resources.js", ["GM_Resources"]);
lazyImport(this, "resource://scriptish/api/GM_setClipboard.js", ["GM_setClipboard"]);

const { add, check } = jetpack('scriptish/security/api-check-filenames');
add(Components.stack.filename);

const NS_XHTML = "http://www.w3.org/1999/xhtml";
const DOLITTLE = function(){};

// Examines the stack to determine if an API should be callable.
function GM_apiLeakCheck(apiName) {
  let stack = Components.stack;

  do {
    // TODO: do better protocol check below
    // Valid stack frames for GM api calls are: native and js when coming from
    // chrome:// URLs and any file name listed in _apiAcceptedFiles.
    if (2 == stack.language && stack.filename &&
        !check(stack.filename) &&
        stack.filename.substr(0, 6) != "chrome") {
      Scriptish_logError(new Error(
          Scriptish_stringBundle("error.api.unsafeAccess") + " - " + apiName + ' by ' + stack.filename));
      return false;
    }
    if (stack.filename && stack.filename === 'resource://scriptish/utils/Scriptish_injectScripts.js') {
      // No point in checking beyond this point.
      // Also fixed at least https://github.com/scriptish/scriptish/issues/231
      break;
    }
  } while (stack = stack.caller);
  return true;
}

function GM_apiSafeCallback(aWindow, aScript, aThis, aCb, aArgs) {
  // Pop back onto browser scope and call event handler.
  // Have to use nested function here otherwise aCallback.apply can point to
  // window.setTimeout, which can be abused to get increased privileges.
  new XPCNativeWrapper(aWindow, "setTimeout()").setTimeout(function() {
    try {
      aCb.apply(aThis, aArgs);
    }
    catch (e) {
      Scriptish_logScriptError(e, aWindow, aScript.fileURL, aScript.id);
    }
  }, 0);
}

// Waive Xrays on an object
const GM_waiveXrays = (function() {
  if (Components.utils.waiveXrays) {
    return function GM_waiveXrays(obj) {
      return Components.utils.waiveXrays(obj);
    };
  }
  return function GM_waiveXrays(obj) {
    return obj;
  };
})();

// Check that o2 has the same principal as o1 and return o2.
// Otherwise return null.
const GM_checkedPrincipal = (function() {
  if (Components.utils.getObjectPrincipal) {
    return function GM_checkedPrincipal(o1, o2) {
      try {
        if (!o2) {
          return null;
        }
        var op1 = Components.utils.getObjectPrincipal(o1);
        var op2 = Components.utils.getObjectPrincipal(o2);
        if (!op1.equals(op2)) {
          Scriptish_logError(new Error("Principal mismatch"));
          return null;
        }
        return o2;
      }
      catch (ex) {
        Scriptish_logError(new Error("Principal mismatch"));
        return null;
      }
    };
  }

  return function GM_checkedPrincipal(o1, o2) {
    return o2;
  }
})();


// note: must not depend on aChromeWin below, it should always be optional!
function GM_API(options) {
  var {
    script: aScript,
    url: aURL,
    winID: aWinID,
    safeWin: aSafeWin,
    unsafeWin: aUnsafeContentWin,
    chromeWin: aChromeWin,
    sandbox: sandbox
  } = options;

  var document = aSafeWin.document;
  var menuCmdIDs = [];
  var Scriptish_BrowserUI = aChromeWin ? aChromeWin.Scriptish_BrowserUI : null;
  var windowID = aWinID;

  var lazyLoaders = {};
  lazy(lazyLoaders, "xhr", function() {
    return new GM_xmlhttpRequester(aUnsafeContentWin, aSafeWin, aURL, aScript, sandbox);
  });
  lazy(lazyLoaders, "storage", function() {
    return new GM_ScriptStorage(aScript, aSafeWin);
  });
  lazy(lazyLoaders, "resources", function() {
    return new GM_Resources(aScript);
  });

  this.GM_safeHTMLParser = function GM_safeHTMLParser(aHTMLStr, aBaseURL) {
    if (!GM_apiLeakCheck("GM_safeHTMLParser")) return;

    let doc = document.implementation.createDocument("", "",
        document.implementation.createDocumentType("html", "", ""));
    doc.appendChild(doc.createElement("html"));
    doc.documentElement.appendChild(doc.createElement("body"));

    let baseURI;
    let frag;

    if ("undefined" !== typeof aBaseURL) {
      try {
        baseURI = NetUtil.newURI(aBaseURL);
      }
      catch(e) {
        throw Error(Scriptish_stringBundle("error.api.safeHTMLParser.url"));
      }
    }
    else {
      baseURI = null;
    }

    // Try to use the newer nsIParserUtils (Gecko >= 14)
    if ("pu" in Services) {
      frag = Services.pu.parseFragment(aHTMLStr, 0, false, baseURI, doc.body);
    }
    // Otherwise fall back to deprecated nsIScriptableUnescapeHTML
    else {
      frag = Services.suhtml.parseFragment(aHTMLStr, false, baseURI, doc.body);
    }
    doc.adoptNode(frag);
    doc.body.appendChild(frag);
    return doc;
  }

  this.GM_notification =
      function GM_notification(aMsg, aTitle, aIcon, aCallback) {
    if (!GM_apiLeakCheck("GM_notification")) return;
    if (typeof aTitle != "string") aTitle = aScript.name;
    if (typeof aIcon != "string") aIcon = aScript.iconURL;

    var callback = null;
    if (typeof aCallback == "function")
      callback = function() GM_apiSafeCallback(aSafeWin, aScript, null, aCallback);
    Scriptish_notification(aMsg, aTitle, aIcon, callback);
  };

  this.GM_setValue = function GM_setValue(aName, aValue) {
    if (!GM_apiLeakCheck("GM_setValue")) return;
    return lazyLoaders.storage.setValue.apply(lazyLoaders.storage, arguments);
  };
  this.GM_getValue = function GM_getValue() {
    if (!GM_apiLeakCheck("GM_getValue")) return;
    return lazyLoaders.storage.getValue.apply(lazyLoaders.storage, arguments);
  };
  this.GM_deleteValue = function GM_deleteValue() {
    if (!GM_apiLeakCheck("GM_deleteValue")) return;
    return lazyLoaders.storage.deleteValue.apply(lazyLoaders.storage, arguments);
  };
  this.GM_watchValue = function GM_watchValue(aName, aListener) {
    if (!GM_apiLeakCheck("GM_watchValue")) return;
    let listener = null;
    if ("function" === typeof aListener) {
      listener = function(aEvent) {
        GM_apiSafeCallback(aSafeWin, aScript, null, aListener, [aEvent]);
      };
    }
    return lazyLoaders.storage.watchValue.call(lazyLoaders.storage, aName, listener);
  }
  this.GM_unwatchValue = function GM_unwatchValue() {
    if (!GM_apiLeakCheck("GM_unwatchValue")) return;
    return lazyLoaders.storage.unwatchValue.apply(lazyLoaders.storage, arguments);
  }
  this.GM_listValues = function GM_listValues() {
    if (!GM_apiLeakCheck("GM_listValues")) return;
    return lazyLoaders.storage.listValues.apply(lazyLoaders.storage, arguments);
  };

  this.GM_getResourceURL = function GM_getResourceURL(aName) {
    if (!GM_apiLeakCheck("GM_getResourceURL")) return;
    return lazyLoaders.resources.getResourceURL.apply(lazyLoaders.resources, arguments)
  }
  this.GM_getResourceText = function GM_getResourceText(aName) {
    if (!GM_apiLeakCheck("GM_getResourceText")) return;
    return lazyLoaders.resources.getResourceText.apply(lazyLoaders.resources, arguments)
  }

  this.GM_getMetadata = function(aKey, aLocalVal) {
    if (!GM_apiLeakCheck("GM_getMetadata")) return;
    return Scriptish_getScriptHeader(aScript, aKey, aLocalVal);
  }

  this.GM_openInTab = function GM_openInTab(aURL, aLoadInBackground, aReuse) {
    if (!GM_apiLeakCheck("GM_openInTab")) return;

    // fix relative links
    aURL = NetUtil.newURI(
        aURL, null, NetUtil.newURI(aSafeWin.location.href, null, null)).spec;

    // open new tab as a child tab of the caller if Tree Style Tab
    // ( https://addons.mozilla.org/firefox/addon/tree-style-tab/ ) there.
    if (aChromeWin.TreeStyleTabService &&
        aChromeWin.TreeStyleTabService.readyToOpenChildTabNow) {
      aChromeWin.TreeStyleTabService.readyToOpenChildTabNow(aSafeWin);
    }

    Scriptish_openInTab(aURL, aLoadInBackground, aReuse, aChromeWin);
    return undefined; // don't return the window, this is intentional
  }

  this.GM_xmlhttpRequest = function GM_xmlhttpRequest() {
    if (!GM_apiLeakCheck("GM_xmlhttpRequest")) return;
    let xhr = lazyLoaders.xhr;
    return xhr.contentStartRequest.apply(xhr, arguments);
  }

  if (!Scriptish_BrowserUI || aSafeWin !== aSafeWin.top) {
    this.GM_unregisterMenuCommand = this.GM_registerMenuCommand
        = this.GM_disableMenuCommand = this.GM_enableMenuCommand = DOLITTLE;
  }
  else {
    this.GM_registerMenuCommand = function GM_registerMenuCommand(
        aCmdName, aCmdFunc, aAccelKey, aAccelModifiers, aAccessKey) {
      if (!GM_apiLeakCheck("GM_registerMenuCommand")) return;

      var uuid = Scriptish_BrowserUI.registerMenuCommand({
        name: aCmdName,
        accelKey: aAccelKey,
        accelModifiers: aAccelModifiers,
        accessKey: aAccessKey,
        doCommand: aCmdFunc,
        winID: windowID});

      menuCmdIDs.push(uuid);

      return uuid;
    }

    this.GM_unregisterMenuCommand = function GM_unregisterMenuCommand(aUUID) {
      var i = menuCmdIDs.indexOf(aUUID);
      if (!~i) return false; // check the uuid is for a cmd made by the same script
      menuCmdIDs.splice(i, 1);
      return Scriptish_BrowserUI.unregisterMenuCommand(aUUID, windowID);
    }

    this.GM_enableMenuCommand = function GM_enableMenuCommand(aUUID) {
      var i = menuCmdIDs.indexOf(aUUID);
      if (!~i) return false; // check the uuid is for a cmd made by the same script
      return Scriptish_BrowserUI.enableMenuCommand(aUUID, windowID);
    }

    this.GM_disableMenuCommand = function GM_disableMenuCommand(aUUID) {
      var i = menuCmdIDs.indexOf(aUUID);
      if (!~i) return false; // check the uuid is for a cmd made by the same script
      return Scriptish_BrowserUI.disableMenuCommand(aUUID, windowID);
    }

    this.GM_addStyle = function GM_addStyle(aSource) {
      if (!GM_apiLeakCheck("GM_addStyle")) return;
      attach(Style({ source: aSource }), aSafeWin);
    }
  }

  this.GM_cryptoHash = function GM_cryptoHash() {
    if (!GM_apiLeakCheck("GM_cryptoHash")) return;
    return Scriptish_cryptoHash.apply(null, arguments);
  }
}

GM_API.prototype.GM_generateUUID = function GM_generateUUID() (
    Services.uuid.generateUUID().toString());

GM_API.prototype.GM_setClipboard = function() {
  if (!GM_apiLeakCheck("GM_setClipboard")) return;
  GM_setClipboard.apply(null, arguments);
}
